"""ILP models module."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pulp import LpContinuous, LpInteger, LpMinimize, LpProblem, LpVariable, lpSum

from qbc_dsbm import QBCHeuristicModels

if TYPE_CHECKING:
    from qbc_dsbm.bigraph import ZerosBiGraph

# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
VertexChoicesLP = list[LpVariable]
EdgeChoicesLP = dict[tuple[int, int], LpVariable]


# ============================================================================ #
#            LP MODEL - DELETING ROWS/COLUMNS FOR ZEROS ELIMINATING            #
# ============================================================================ #
def min_del_rc(
    bigraph: ZerosBiGraph,
    epsilon: float,
) -> tuple[LpProblem, VertexChoicesLP, VertexChoicesLP, EdgeChoicesLP]:
    """Minimizing the number of deleted rows/columns.

    Parameters
    ----------
    bigraph : BiGraph
        Bigraph
    epsilon : float
        Error rate

    Returns
    -------
    LpProblem
        PuLP problem
    VertexChoicesLP
        U vertex choices
    VertexChoicesLP
        V vertex choices
    EdgeChoicesLP
        Edge choices

    """
    prob = LpProblem(name=QBCHeuristicModels.MIN_DEL_RC.value, sense=LpMinimize)
    #
    # Variables
    #
    u_vertex_choices = __u_vertex_choices_variables(bigraph)
    v_vertex_choices = __v_vertex_choices_variables(bigraph)
    edge_choices = __edge_choices_variables(bigraph)
    #
    # Objective function
    #
    prob += lpSum(u_vertex_choices) + lpSum(v_vertex_choices), "min_vertices"
    #
    # Constraints
    #
    __select_edge_select_at_least_one_endpoint(
        prob,
        bigraph,
        u_vertex_choices,
        v_vertex_choices,
        edge_choices,
    )
    __at_most_epsilon_remaining_zeros(prob, bigraph, epsilon, edge_choices)

    return prob, u_vertex_choices, v_vertex_choices, edge_choices


def min_del_rows(
    bigraph: ZerosBiGraph,
    epsilon: float,
) -> tuple[LpProblem, VertexChoicesLP, EdgeChoicesLP]:
    """Minimizing the number of deleted rows.

    Parameters
    ----------
    bigraph : BiGraph
        Bigraph
    epsilon : float
        Error rate

    Returns
    -------
    LpProblem
        PuLP problem
    VertexChoicesLP
        U vertex choices
    EdgeChoicesLP
        Edge choices

    """
    prob = LpProblem(name=QBCHeuristicModels.MIN_DEL_ROWS.value, sense=LpMinimize)
    #
    # Variables
    #
    u_vertex_choices = __u_vertex_choices_variables(bigraph)
    edge_choices = __edge_choices_variables(bigraph)
    #
    # Objective function
    #
    prob += lpSum(u_vertex_choices), "min_u_vertices"
    #
    # Constraints
    #
    __select_edge_select_u_endpoint(prob, bigraph, u_vertex_choices, edge_choices)
    __at_most_epsilon_remaining_zeros(prob, bigraph, epsilon, edge_choices)

    return prob, u_vertex_choices, edge_choices


def min_del_cols(
    bigraph: ZerosBiGraph,
    epsilon: float,
) -> tuple[LpProblem, VertexChoicesLP, EdgeChoicesLP]:
    """Minimizing the number of deleted rows.

    Parameters
    ----------
    bigraph : BiGraph
        Bigraph
    epsilon : float
        Error rate

    Returns
    -------
    LpProblem
        PuLP problem
    VertexChoicesLP
        V vertex choices
    EdgeChoicesLP
        Edge choices

    """
    prob = LpProblem(name=QBCHeuristicModels.MIN_DEL_COLS.value, sense=LpMinimize)
    #
    # Variables
    #
    v_vertex_choices = __v_vertex_choices_variables(bigraph)
    edge_choices = __edge_choices_variables(bigraph)
    #
    # Objective function
    #
    prob += lpSum(v_vertex_choices), "min_v_vertices"
    #
    # Constraints
    #
    __select_edge_select_v_endpoint(prob, bigraph, v_vertex_choices, edge_choices)
    __at_most_epsilon_remaining_zeros(prob, bigraph, epsilon, edge_choices)

    return prob, v_vertex_choices, edge_choices


def min_del_rows_relax(
    bigraph: ZerosBiGraph,
    epsilon: float,
) -> tuple[LpProblem, VertexChoicesLP, EdgeChoicesLP]:
    """Minimizing the number of deleted rows.

    Parameters
    ----------
    bigraph : BiGraph
        Bigraph
    epsilon : float
        Error rate

    Returns
    -------
    LpProblem
        PuLP problem
    VertexChoicesLP
        U vertex choices
    EdgeChoicesLP
        Edge choices

    """
    prob = LpProblem(name=QBCHeuristicModels.MIN_DEL_ROWS_RELAX.value, sense=LpMinimize)
    #
    # Variables
    #
    u_vertex_choices = __u_vertex_choices_variables(bigraph, var_type=LpContinuous)
    edge_choices = __edge_choices_variables(bigraph)
    #
    # Objective function
    #
    prob += lpSum(u_vertex_choices), "min_u_vertices"
    #
    # Constraints
    #
    __select_edge_select_u_endpoint(prob, bigraph, u_vertex_choices, edge_choices)
    __at_most_epsilon_remaining_zeros(prob, bigraph, epsilon, edge_choices)

    return prob, u_vertex_choices, edge_choices


def min_del_cols_relax(
    bigraph: ZerosBiGraph,
    epsilon: float,
) -> tuple[LpProblem, VertexChoicesLP, EdgeChoicesLP]:
    """Minimizing the number of deleted rows.

    Parameters
    ----------
    bigraph : BiGraph
        Bigraph
    epsilon : float
        Error rate

    Returns
    -------
    LpProblem
        PuLP problem
    VertexChoicesLP
        V vertex choices
    EdgeChoicesLP
        Edge choices

    """
    prob = LpProblem(name=QBCHeuristicModels.MIN_DEL_COLS_RELAX.value, sense=LpMinimize)
    #
    # Variables
    #
    v_vertex_choices = __v_vertex_choices_variables(bigraph, var_type=LpContinuous)
    edge_choices = __edge_choices_variables(bigraph)
    #
    # Objective function
    #
    prob += lpSum(v_vertex_choices), "min_v_vertices"
    #
    # Constraints
    #
    __select_edge_select_v_endpoint(prob, bigraph, v_vertex_choices, edge_choices)
    __at_most_epsilon_remaining_zeros(prob, bigraph, epsilon, edge_choices)

    return prob, v_vertex_choices, edge_choices


def min_del_ones(
    bigraph: ZerosBiGraph,
    epsilon: float,
) -> tuple[LpProblem, VertexChoicesLP, VertexChoicesLP, EdgeChoicesLP]:
    """Minimizing the number of deleted ones.

    Parameters
    ----------
    bigraph : BiGraph
        Bigraph
    epsilon : float
        Error rate

    Returns
    -------
    LpProblem
        PuLP problem
    VertexChoicesLP
        U vertex choices
    VertexChoicesLP
        V vertex choices
    EdgeChoicesLP
        Edge choices

    """
    prob = LpProblem(name=QBCHeuristicModels.MIN_DEL_ONES.value, sense=LpMinimize)
    #
    # Variables
    #
    u_vertex_choices = __u_vertex_choices_variables(bigraph)
    v_vertex_choices = __v_vertex_choices_variables(bigraph)
    edge_choices = __edge_choices_variables(bigraph)
    #
    # Objective function
    #
    prob += (
        (
            lpSum(
                (
                    var * deg / bigraph.card_v_bipart()
                    for var, deg in zip(
                        u_vertex_choices,
                        bigraph.u_bipart_anti_degrees(),
                    )
                ),
            )
            + lpSum(
                (
                    var * deg / bigraph.card_u_bipart()
                    for var, deg in zip(
                        v_vertex_choices,
                        bigraph.v_bipart_anti_degrees(),
                    )
                ),
            )
        ),
        "min_weighted_rows/cols",
    )
    #
    # Constraints
    #
    __select_edge_select_at_least_one_endpoint(
        prob,
        bigraph,
        u_vertex_choices,
        v_vertex_choices,
        edge_choices,
    )
    __at_most_epsilon_remaining_zeros(prob, bigraph, epsilon, edge_choices)

    return prob, u_vertex_choices, v_vertex_choices, edge_choices


# ============================================================================ #
#                           LP MODEL - KNAPSACK MODEL                          #
# ============================================================================ #
def kp_qb(
    bigraph: ZerosBiGraph,
    epsilon: float,
) -> tuple[LpProblem, VertexChoicesLP, VertexChoicesLP]:
    """Knapsack model with zero degrees uncertainty.

    Parameters
    ----------
    bigraph : BiGraph
        Bigraph
    epsilon : float
        Error rate

    Returns
    -------
    LpProblem
        PuLP problem
    VertexChoicesLP
        U vertex choices
    VertexChoicesLP
        V vertex choices

    """
    prob = LpProblem(name=QBCHeuristicModels.KP_QB.value, sense=LpMinimize)
    #
    # Variables
    #
    u_vertex_choices = __u_vertex_choices_variables(bigraph)
    v_vertex_choices = __v_vertex_choices_variables(bigraph)
    #
    # Objective function
    #
    prob += (
        (
            lpSum(
                (
                    var * deg / bigraph.card_v_bipart()
                    for var, deg in zip(
                        u_vertex_choices,
                        bigraph.u_bipart_anti_degrees(),
                    )
                ),
            )
            + lpSum(
                (
                    var * deg / bigraph.card_u_bipart()
                    for var, deg in zip(
                        v_vertex_choices,
                        bigraph.v_bipart_anti_degrees(),
                    )
                ),
            )
        ),
        "knapsack",
    )
    #
    # Constraints
    #
    __at_most_epsilon_zero_degrees(
        prob,
        bigraph,
        epsilon,
        u_vertex_choices,
        v_vertex_choices,
    )

    return prob, u_vertex_choices, v_vertex_choices


# ============================================================================ #
#                                   VARIABLES                                  #
# ============================================================================ #
def __u_vertex_choices_variables(
    bigraph: ZerosBiGraph,
    var_type: str = LpInteger,
) -> VertexChoicesLP:
    return [
        LpVariable(f"xu_{u}", cat=var_type, lowBound=0, upBound=1)
        for u in range(bigraph.card_u_bipart())
    ]


def __v_vertex_choices_variables(
    bigraph: ZerosBiGraph,
    var_type: str = LpInteger,
) -> VertexChoicesLP:
    return [
        LpVariable(f"xv_{v}", cat=var_type, lowBound=0, upBound=1)
        for v in range(bigraph.card_v_bipart())
    ]


def __edge_choices_variables(
    bigraph: ZerosBiGraph,
    var_type: str = LpInteger,
) -> EdgeChoicesLP:
    return {
        (u, v): LpVariable(f"edge_{u}_{v}", cat=var_type, lowBound=0, upBound=1)
        for u, v in bigraph.edges()
    }


# ============================================================================ #
#                                  CONSTRAINTS                                 #
# ============================================================================ #
def __select_edge_select_at_least_one_endpoint(
    prob: LpProblem,
    bigraph: ZerosBiGraph,
    u_vertex_choices: VertexChoicesLP,
    v_vertex_choices: VertexChoicesLP,
    edge_choices: EdgeChoicesLP,
) -> None:
    for u, v in bigraph.edges():
        prob += (
            (u_vertex_choices[u] + v_vertex_choices[v] >= edge_choices[u, v]),
            f"edge_{u}_{v}",
        )


def __select_edge_select_u_endpoint(
    prob: LpProblem,
    bigraph: ZerosBiGraph,
    u_vertex_choices: VertexChoicesLP,
    edge_choices: EdgeChoicesLP,
) -> None:
    for u, v in bigraph.edges():
        prob += (u_vertex_choices[u] >= edge_choices[u, v]), f"edge_{u}_{v}"


def __select_edge_select_v_endpoint(
    prob: LpProblem,
    bigraph: ZerosBiGraph,
    v_vertex_choices: VertexChoicesLP,
    edge_choices: EdgeChoicesLP,
) -> None:
    for u, v in bigraph.edges():
        prob += (v_vertex_choices[v] >= edge_choices[u, v]), f"edge_{u}_{v}"


def __at_most_epsilon_remaining_zeros(
    prob: LpProblem,
    bigraph: ZerosBiGraph,
    epsilon: float,
    edge_choices: EdgeChoicesLP,
) -> None:
    prob += (lpSum(edge_choices) >= (1 - epsilon) * len(bigraph.edges())), "sensitivity"


def __at_most_epsilon_zero_degrees(
    prob: LpProblem,
    bigraph: ZerosBiGraph,
    epsilon: float,
    u_vertex_choices: VertexChoicesLP,
    v_vertex_choices: VertexChoicesLP,
) -> None:
    prob += (
        (
            lpSum(
                (
                    var * deg
                    for var, deg in zip(u_vertex_choices, bigraph.u_bipart_degrees())
                ),
            )
            + lpSum(
                (
                    var * deg
                    for var, deg in zip(v_vertex_choices, bigraph.v_bipart_degrees())
                ),
            )
            >= (1 - epsilon) * len(bigraph.edges())
        ),
        "sensitivity",
    )



"""Binary matrix views."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from qbc_dsbm.binary_matrix import BinMatrix


class BinMatrixStats:
    """Binary matrix stats."""

    __NUMBER_OF_ROWS_KEY: str = "number_of_rows"
    __NUMBER_OF_COLUMNS_KEY: str = "number_of_columns"
    __NUMBER_OF_ONES_KEY: str = "number_of_ones"
    __NUMBER_OF_ZEROS_KEY: str = "number_of_zeros"
    __DENSITY_KEY: str = "density"
    __SPARSITY_KEY: str = "sparsity"

    @classmethod
    def from_dict(cls, bin_matrix_stats_dict: dict[str, Any]) -> BinMatrixStats:
        """Instantiate binary matrix stats from a dictionnary.

        Parameters
        ----------
        bin_matrix_stats_dict : dict
            Binary matrix stats dictionnary

        Returns
        -------
        BinMatrixStats
            Binary matrix stats

        """
        return BinMatrixStats(
            bin_matrix_stats_dict[cls.__NUMBER_OF_ROWS_KEY],
            bin_matrix_stats_dict[cls.__NUMBER_OF_COLUMNS_KEY],
            bin_matrix_stats_dict[cls.__NUMBER_OF_ONES_KEY],
            bin_matrix_stats_dict[cls.__NUMBER_OF_ZEROS_KEY],
            bin_matrix_stats_dict[cls.__DENSITY_KEY],
            bin_matrix_stats_dict[cls.__SPARSITY_KEY],
        )

    @classmethod
    def from_bin_matrix(cls, bin_matrix: BinMatrix) -> BinMatrixStats:
        """Instantiate a binary matrix stats from a binary matrix.

        Parameters
        ----------
        bin_matrix : BinMatrix
            Binary matrix

        Returns
        -------
        BinMatrixStats
            Binary matrix stats

        """
        nb_cells = bin_matrix.number_of_rows() * bin_matrix.number_of_columns()
        return cls(
            bin_matrix.number_of_rows(),
            bin_matrix.number_of_columns(),
            bin_matrix.number_of_ones(),
            bin_matrix.number_of_zeros(),
            bin_matrix.number_of_ones() / nb_cells,
            bin_matrix.number_of_zeros() / nb_cells,
        )

    def __init__(  # noqa: PLR0913
        self,
        number_of_rows: int,
        number_of_columns: int,
        number_of_ones: int,
        number_of_zeros: int,
        density: float,
        sparsity: float,
    ) -> None:
        """Initialize binary matrix stats."""
        self.__number_of_rows = number_of_rows
        self.__number_of_columns = number_of_columns
        self.__number_of_ones = number_of_ones
        self.__number_of_zeros = number_of_zeros
        self.__density = density
        self.__sparsity = sparsity

    def number_of_rows(self) -> int:
        """Give the number of rows.

        Returns
        -------
        int
            Number of rows

        """
        return self.__number_of_rows

    def number_of_columns(self) -> int:
        """Give the number of columns.

        Returns
        -------
        int
            Number of columns

        """
        return self.__number_of_columns

    def number_of_ones(self) -> int:
        """Give the number of ones.

        Returns
        -------
        int
            Number of ones

        """
        return self.__number_of_ones

    def number_of_zeros(self) -> int:
        """Give the number of zeros.

        Returns
        -------
        int
            Number of zeros

        """
        return self.__number_of_zeros

    def density(self) -> float:
        """Density.

        Returns
        -------
        float
            Density

        """
        return self.__density

    def sparsity(self) -> float:
        """Sparsity.

        Returns
        -------
        float
            Sparsity

        """
        return self.__sparsity

    def to_dict(self) -> dict:
        """Convert binary matrix stats to dict.

        Returns
        -------
        dict
            Dictionnary

        """
        return {
            self.__NUMBER_OF_ROWS_KEY: self.__number_of_rows,
            self.__NUMBER_OF_COLUMNS_KEY: self.__number_of_columns,
            self.__NUMBER_OF_ONES_KEY: self.__number_of_ones,
            self.__NUMBER_OF_ZEROS_KEY: self.__number_of_zeros,
            self.__DENSITY_KEY: self.__density,
            self.__SPARSITY_KEY: self.__sparsity,
        }

    def __str__(self) -> str:
        """Print stats."""
        return "\n".join(
            [
                f"Dimensions: {self.number_of_rows()} x {self.number_of_columns()}"
                f" ({self.number_of_rows() * self.number_of_columns()})",
                f"Number of ones: {self.number_of_ones()}",
                f"Number of zeros: {self.number_of_zeros()}",
                f"Density: {self.density()}",
                f"Sparsity: {self.sparsity()}",
            ],
        )



"""Binary matrix module."""

from qbc_dsbm.binary_matrix.configs import KNNCompletetionConfig
from qbc_dsbm.binary_matrix.create import (
    complete_dataframe_with_knn,
    encapsulate_permutate_bin_dataframe,
    generate_bin_dataframe,
    generate_bin_matrix,
)
from qbc_dsbm.binary_matrix.items import (
    BinMatrix,
    DegeneratedMatrixError,
    NotABinaryMatrixError,
)
from qbc_dsbm.binary_matrix.ops import (
    a_inter_b_columns,
    a_inter_b_rows,
    a_minus_b_columns,
    a_minus_b_rows,
)
from qbc_dsbm.binary_matrix.views import BinMatrixStats

__all__ = [
    "BinMatrix",
    "DegeneratedMatrixError",
    "NotABinaryMatrixError",
    "BinMatrixStats",
    "a_inter_b_columns",
    "a_inter_b_rows",
    "a_minus_b_columns",
    "a_minus_b_rows",
    "generate_bin_matrix",
    "generate_bin_dataframe",
    "encapsulate_permutate_bin_dataframe",
    "KNNCompletetionConfig",
    "complete_dataframe_with_knn",
]



"""Application module for binary matrix creation."""

# Due to typer usage:
# ruff: noqa: TCH001, TCH003, UP007, FBT001, FBT002, PLR0913

from __future__ import annotations

import ast
import logging
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Optional, cast

import numpy as np
import pandas as pd
import typer

import qbc_dsbm.args as common_args
import qbc_dsbm.binary_matrix.configs as bm_cfg
import qbc_dsbm.binary_matrix.create as bm_create
import qbc_dsbm.logging as common_log
from qbc_dsbm.binary_matrix.items import BinMatrix, NotABinaryMatrixError

_LOGGER = logging.getLogger(__name__)

APP = typer.Typer(rich_markup_mode="rich")


@dataclass
class CommonArgs:
    """Common args and opts."""

    ARG_INPUT_MATRIX = typer.Argument(
        help="Path to the non-binary matrix",
    )

    ARG_OUTPUT_MATRIX = typer.Argument(
        help="Path to the output binary matrix",
    )


@dataclass
class CustomArgs:
    """Custom args and opts."""

    ARG_VALUE_REPLACED = typer.Argument(
        help=('Value to replace as a python dict in a string, e.g. "{-1: 1, -2: 0}"'),
    )


@APP.command()
def custom(
    input_matrix: Path = CommonArgs.ARG_INPUT_MATRIX,
    output_matrix: Path = CommonArgs.ARG_OUTPUT_MATRIX,
    value_replaced: str = CustomArgs.ARG_VALUE_REPLACED,
) -> None:
    """Transform non-binary matrix to binary matrix with custom replacement."""
    try:
        value_replaced_dict = ast.literal_eval(value_replaced)
    except SyntaxError:
        _LOGGER.critical("Invalid value_replaced: %s", value_replaced_dict)
        sys.exit(1)
    if not isinstance(value_replaced_dict, dict):
        _LOGGER.critical("Invalid value_replaced: %s", value_replaced_dict)
        sys.exit(1)

    try:
        bin_matrix = BinMatrix.from_csv(
            input_matrix,
            replace_value_map=cast(dict, value_replaced),
        )
    except NotABinaryMatrixError as err:
        _LOGGER.critical("Output matrix is not a binary matrix:")
        _LOGGER.critical(err)
        sys.exit(1)

    bin_matrix.to_csv(output_matrix)


@APP.command()
def random(
    input_matrix: Path = CommonArgs.ARG_INPUT_MATRIX,
    output_matrix: Path = CommonArgs.ARG_OUTPUT_MATRIX,
) -> None:
    """Transform non-binary matrix to binary matrix with random replacement."""
    df: pd.DataFrame = pd.read_csv(input_matrix, header=0, index_col=0)

    rng = np.random.default_rng()
    for i in range(df.shape[0]):
        for j in range(df.shape[1]):
            if df.iloc[i, j] not in (0, 1):
                df.iloc[i, j] = rng.choice([0, 1], 1)

    BinMatrix.new_unsafe(df).to_csv(output_matrix)


@dataclass
class KNNArgs:
    """KNN args and opts."""

    __CFG_CAT = "KNN configuration"

    OPT_NUMBER_OF_NEIGHBOURS = typer.Option(
        "--number-of-neighbours",
        "-n",
        help="Number of nearest neighbours",
        rich_help_panel=__CFG_CAT,
    )

    OPT_THRESHOLD = typer.Option(
        "--threshold",
        "-t",
        help=(
            "Threshold confident interval:"
            " if 0 <= values < threshold, then value -> 0,"
            " else if threshold <= values <= 1, then value -> 1,"
            " else value -> undefined"
        ),
        rich_help_panel=__CFG_CAT,
    )

    OPT_CONFIG_FILE = typer.Option(
        "--config",
        help="Path to the config file",
        rich_help_panel=__CFG_CAT,
    )


@APP.command()
def knn(
    input_matrix: Annotated[Path, CommonArgs.ARG_INPUT_MATRIX],
    output_matrix: Annotated[Path, CommonArgs.ARG_OUTPUT_MATRIX],
    number_of_neighbours: Annotated[
        int,
        KNNArgs.OPT_NUMBER_OF_NEIGHBOURS,
    ] = bm_cfg.KNNCompletetionConfig.DEFAULT_NUMBER_OF_NEIGHBOURS,
    threshold: Annotated[
        float,
        KNNArgs.OPT_THRESHOLD,
    ] = bm_cfg.KNNCompletetionConfig.DEFAULT_THRESHOLD,
    config_file: Annotated[
        Optional[Path],
        KNNArgs.OPT_CONFIG_FILE,
    ] = None,
    debug: Annotated[
        bool,
        common_args.CommonArgs.OPT_DEBUG,
    ] = False,
) -> None:
    """Transform non-binary matrix to binary matrix with KNN replacement."""
    common_log.init_logger("KNN replacement command", debug)
    _LOGGER.info("Input matrix: %s", input_matrix)
    _LOGGER.info("Output matrix: %s", output_matrix)

    df: pd.DataFrame = pd.read_csv(input_matrix, header=0, index_col=0)

    knn_config = (
        bm_cfg.KNNCompletetionConfig.from_yaml(config_file)
        if config_file is not None
        else bm_cfg.KNNCompletetionConfig(
            number_of_neighbours=number_of_neighbours,
            threshold=threshold,
        )
    )

    try:
        bin_matrix = bm_create.complete_dataframe_with_knn(df, knn_config)
    except NotABinaryMatrixError as err:
        _LOGGER.critical("Output matrix is not a binary matrix:")
        _LOGGER.critical(err)
        sys.exit(1)

    bin_matrix.to_csv(output_matrix)


if __name__ == "__main__":
    APP()



"""Binary matrix configs module."""

from __future__ import annotations

from pathlib import Path

import yaml  # type: ignore[import-untyped]

try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper


class KNNCompletetionConfig:
    """KNN completetion configuration."""

    DEFAULT_NUMBER_OF_NEIGHBOURS = 5
    DEFAULT_THRESHOLD = 0.3

    KEY_NUMBER_OF_NEIGHBOURS = "number_of_neighbours"
    KEY_THRESHOLD = "threshold"

    DEFAULT_YAML_FILE = Path("knn_completetion_config.yaml")

    @classmethod
    def default_yaml_filepath(cls, parent_dir: Path) -> Path:
        """Return the default YAML file path."""
        return parent_dir / cls.DEFAULT_YAML_FILE

    @classmethod
    def from_dir(cls, parent_dir: Path) -> KNNCompletetionConfig:
        """Instantiate from directory."""
        return cls.from_yaml(cls.default_yaml_filepath(parent_dir))

    @classmethod
    def from_yaml(cls, yaml_filepath: Path) -> KNNCompletetionConfig:
        """Instantiate from YAML."""
        with yaml_filepath.open(encoding="utf-8") as yaml_file:
            yaml_dict = yaml.safe_load(yaml_file)
            return cls.from_dict(yaml_dict)

    @classmethod
    def from_dict(cls, yaml_dict: dict) -> KNNCompletetionConfig:
        """Instantiate from dictionary."""
        return cls(
            number_of_neighbours=yaml_dict.get(
                cls.KEY_NUMBER_OF_NEIGHBOURS,
                cls.DEFAULT_NUMBER_OF_NEIGHBOURS,
            ),
            threshold=yaml_dict.get(
                cls.KEY_THRESHOLD,
                cls.DEFAULT_THRESHOLD,
            ),
        )

    def __init__(
        self,
        number_of_neighbours: int = DEFAULT_NUMBER_OF_NEIGHBOURS,
        threshold: float = DEFAULT_THRESHOLD,
    ) -> None:
        """Initialize."""
        self.__number_of_neighbours = number_of_neighbours
        self.__threshold = threshold

    def number_of_neighbours(self) -> int:
        """Return the number of neighbours."""
        return self.__number_of_neighbours

    def threshold(self) -> float:
        """Return the threshold."""
        return self.__threshold

    def to_dict(self) -> dict:
        """Return the dictionary representation."""
        return {
            self.KEY_NUMBER_OF_NEIGHBOURS: self.__number_of_neighbours,
            self.KEY_THRESHOLD: self.__threshold,
        }

    def to_yaml(self, yaml_filepath: Path) -> None:
        """Write to YAML."""
        with yaml_filepath.open("w", encoding="utf-8") as yaml_file:
            yaml.dump(self.to_dict(), yaml_file, Dumper=Dumper, sort_keys=False)

    def to_dir(self, parent_dir: Path) -> None:
        """Write to directory."""
        self.to_yaml(self.default_yaml_filepath(parent_dir))

    def __str__(self) -> str:
        """Return the string representation."""
        return "\n".join(
            [
                f"Number of neighbours: {self.__number_of_neighbours}",
                f"Threshold: {self.__threshold} such as"
                f" [0, {self.__threshold}[ => 0, [{self.__threshold}, 1] => 1",
            ],
        )
